% FGHC -> FLENG compiler (baseline version)

:- dynamic([declaration/1, library_mode/0]).
:- initialization(main).

:- if(current_prolog_flag(dialect, gprolog)).
% GNU
:- include('version.pl').
main :-
    argument_list(Args),
    main(Args).
:- endif.

:- if(current_prolog_flag(dialect, swi)).
% SWI
:- include(fleng('version.pl')).

format_to_atom(Out, Fstr, Args) :-
    format_to_codes(Fstr, Args, S),
    atom_codes(Out, S).

g_assign(Var, Val) :- nb_setval(Var, Val).
g_read(Var, Val) :- nb_getval(Var, Val).
:- endif.

:- if(current_prolog_flag(dialect, yap)).
% YAP
:- use_module(library(lists)).
:- use_module(library(charsio)).
:- include('version.pl').

format_to_atom(Out, Fstr, Args) :-
    format_to_chars(Fstr, Args, S),
    atom_codes(Out, S).

save_to_file :- save('fghc2fl.yap'), halt(0).

start :- 
    set_prolog_flag(syntax_errors, error),
    set_prolog_flag(unknown, error),
    unix(argv(Args)), 
    catch(main(Args), E, quit(E)).

quit(E) :- 
    print_message(error, E),
    halt(1).

numbervars(Term) :- numbervars(Term, 0, _).

g_assign(Var, Val) :- nb_setval(Var, Val).
g_read(Var, Val) :- nb_getval(Var, Val).
:- endif.

main(['-v']) :-
    fleng_version(V),
    format('fghc2fl ~a\n', [V]),
    halt(0).
main([Infile, Outfile]) :-
    op(200, fx, !),
    op(700, xfx, '\\=:='),
    op(500, yfx, '><'),
    op(400, yfx, '\\\\'),
    op(700, yfx, '@'),
    op(0, fx, initialization),
    set_prolog_flag(double_quotes, codes),
    g_assign(counter, 1),
    see(Infile),
    read_program(Prg),
    seen,
    collect_groups(Prg, Groups),
    translate(Groups, TG),
    open(Outfile, write, Out),
    fleng_version(V),
    format(Out, '%% GENERATED BY fghc2fl [~a] FROM ~a\n\n', [V, Infile]),
    write_declarations(Out),
    write_translation(TG, Out),
    format(Out, '%% END OF FILE\n', []),
    close(Out),
    halt(0).
main(_) :- 
    format(user_error, 'usage: fghc2fl [-v] INFILE OUTFILE\n', []),
    halt(1).

dbg(X) :- format('~q\n', [X]).
dbg(F, X) :- format(F, [X]).

error(Fstr, Args) :-
    format(user_error, Fstr, Args),
    halt(1).

read_program(PRG) :-
    read_term(TERM, [singletons(SL)]),
    report_singletons(SL, TERM),
    numbervars(TERM),
    read_program(TERM, PRG).

read_program(end_of_file, []).
read_program(-DECL, PRG) :-
    process_declaration(DECL, PRG).
read_program(TERM, [c(NA, Args, Guards, Body)|PRG]) :-
    canonicalize(TERM, NA, Args, Guards, Body),
    read_program(PRG).

report_singletons([], _).
report_singletons(SL, _) :-
    format(user_error, 'Warning: singleton variables ', []),
    member(Name = _, SL),
    format(user_error, '~a ', [Name]),
    fail.
report_singletons(_, Term) :-
    format(user_error, 'in term:\n\n  ~q\n', [Term]).

process_declaration(library_mode, PRG) :-
    assertz(library_mode),
    read_program(PRG).
process_declaration(include(Fname), PRG) :-
    current_input(In),
    see(Fname),
    read_program(Included),
    seen,
    set_input(In),
    append(Included, Next, PRG),
    read_program(Next).
process_declaration(Decl, PRG) :-
    assertz(declaration(Decl)),
    read_program(PRG).

write_declarations(Out) :-
    declaration(Decl),
    format(Out, '-~q.\n', [Decl]),
    fail.
write_declarations(Out) :- 
    \+library_mode,
    format(Out, '-uses([lib, sys]).\n', []),
    fail.
write_declarations(Out) :- nl(Out).

collect_groups([], []).
collect_groups([T|PRG], [[T|G]|GROUPS]) :-
    T = c(NA, _, _, _),
    filter_group(PRG, NA, NA, G, NG),
    collect_groups(NG, GROUPS).

filter_group([], _, _, [], []).
filter_group([T|PRG], NA, NA, [T|G], NG) :-
    T = c(NA, _, _, _),
    !,
    filter_group(PRG, NA, NA, G, NG).
filter_group([T|PRG], N/A, _, [T|G], NG) :-
    T = c(N/A, _, _, _),
    !,
    format(user_error, 'Warning: non-contiguous process definition for ~a/~d\n', [N, A]),
    filter_group(PRG, NA, NA, G, NG).
filter_group([T|PRG], NA, _, G, [T|NG]) :-
    T = c(NA2, _, _, _),    
    filter_group(PRG, NA, NA2, G, NG).

canonicalize((H :- G | B), N/A, Args, Guards, B) :-
    H =.. [N|Args],
    length(Args, A),
    serialize_guards(G, Guards).
canonicalize((H :- B), N/A, Args, Guards, B) :-
    canonicalize((H :- true | B), N/A, Args, Guards, B).
canonicalize(H, N/A, Args, Guards, B) :-
    canonicalize((H :- true | true), N/A, Args, Guards, B).

serialize_guards(true, []).
serialize_guards((true, G1), G) :- serialize_guards(G1, G).
serialize_guards((G1, G2), [G1|G3]) :- serialize_guards(G2, G3).
serialize_guards(G, [G]).

write_translation([], _).
write_translation([[]|Gs], Out) :-
    write_translation(Gs, Out).
write_translation([G|Gs], Out) :-
    write_group(G, Out),
    nl(Out),
    write_translation(Gs, Out).

write_group([], _).
write_group([c(N/_, Args, _, B)|G], Out) :-
    H =.. [N|Args],
    format(Out, '~q.\n', [(H :- B)]),
    write_group(G, Out).

% translation chain

translate(G, TG) :-
    check_definitions(G),
    inject_equality_guards(G, G2),
    expand_terms(G2, G3),
    eliminate_guards(G3, TG).

% check for library redefinitions

check_definitions(_) :- library_mode, !.
check_definitions(Gs) :-
    member(G, Gs), 
    check_definitions2(G), 
    fail.
check_definitions(_).

check_definitions2([c(N/A, _, _, _)|_]) :-
    \+foreign(N/A, _, _),
    \+library(N/A, _),
    !.
check_definitions2([c(N/A, _, _, _)|_]) :-
    error('redefinition of library predicate ~a/~d\n', [N, A]).

% inject ==/2 guards for variables occurring multiple times in head

inject_equality_guards([], []).
inject_equality_guards([G1|Gs1], [G|Gs]) :-
    inject_equality_guards2(G1, G),
    inject_equality_guards(Gs1, Gs).

inject_equality_guards2([], []).
inject_equality_guards2([c(NA, Args1, Guards1, Body)|G1], [T|G]) :-
    T = c(NA, Args, Guards, Body),
    extract_variables((Args1, Body), Vars),
    duplicated_variables(Args1, Args, [], _, Vars, _, Guards1, Guards),
    inject_equality_guards2(G1, G).

duplicated_variables(X, X, S, S, Vs, Vs, G, G) :- atomic(X).
duplicated_variables('$VAR'(V), '$VAR'(V2), S, S, Vs, [V2|Vs], G,  [G1|G]) :-
    member(V, S),
    available_variable(Vs, V2),
    G1 = ('$VAR'(V) == '$VAR'(V2)).
duplicated_variables('$VAR'(V), '$VAR'(V), S, [V|S], Vs, Vs, G, G).
duplicated_variables([X1|Y1], [X|Y], S1, S, Vs1, Vs, G1, G) :-
    duplicated_variables(X1, X, S1, S2, Vs1, Vs2, G1, G2),
    duplicated_variables(Y1, Y, S2, S, Vs2, Vs, G2, G).
duplicated_variables(T1, T, S1, S, Vs1, Vs, G1, G) :-
    callable(T1),
    T1 =.. [Name|Args1],
    duplicated_variables(Args1, Args, S1, S, Vs1, Vs, G1, G),
    T =.. [Name|Args].
duplicated_variables(X, X, S, S, Vs, Vs, G, G).

% eliminate guards by introducing new groups

eliminate_guards([], []).
eliminate_guards([G1|Gs1], [G|Gs]) :-
    eliminate_guards(G1, G, Gs1, Gs2),
    eliminate_guards(Gs2, Gs).

eliminate_guards([c(NA, Args, Guards1, B)|G1], G, Gs1, Gs) :-
    % 'otherwise' in first clause has no effect
    delete(Guards1, otherwise, Guards),
    eliminate_guards2([c(NA, Args, Guards, B)|G1], G, Gs1, Gs).

eliminate_guards2([], [], Gs, Gs).
eliminate_guards2([C|Cs1], [C|Cs], Gs1, Gs) :-
    C = c(_, _, [], _),
    % no guards, try next clause in group
    eliminate_guards2(Cs1, Cs, Gs1, Gs).
eliminate_guards2([C1|Cs1], Cs, Gs1, Gs) :-
    C1 = c(NA, Args, Guards, B),
    member(otherwise, Guards),
    eliminate_guards_otherwise([c(NA, Args, Guards, B)|Cs1], Cs, Gs1, Gs).
eliminate_guards2([C1|Cs1], Cs, Gs1, Gs) :-
    C1 = c(NA, Args, [idle|Guards], B),
    % idle/0 guard
    eliminate_guards_idle(NA, Args, Guards, B, Cs1, Cs, Gs1, Gs).
eliminate_guards2([C1|Cs1], Cs, Gs1, Gs) :-
    C1 = c(NA, Args, [data('$VAR'(V))|Guards], B),
    % data/1 guard
    force_variables(Args, [V], Args2),
    eliminate_guards2([c(NA, Args2, Guards, B)|Cs1], Cs, Gs1, Gs).
eliminate_guards2([C1], [C], Gs1, Gs) :-
    % final clause with guards, excluding 'otherwise'
    C1 = c(NA, Args, Guards, B),
    C = c(NA, Args2, [], B2),
    eliminate_guards(Guards, NA, Args, Args2, B, B2, [], Gs1, Gs).
eliminate_guards2([C1|Cs1], [C], Gs1, [Cs|Gs]) :-
    % non-final clause with guards, excluding 'otherwise'
    C1 = c(NA, Args, Guards, B),
    C = c(NA, Args2, [], B2),
    split_group(Cs1, 1, Cs),
    rename_functor(NA, 1, Next),
    eliminate_guards(Guards, NA, Args, Args2, B, B2, Next, Gs1, Gs).

eliminate_guards_otherwise([c(NA, Args, Guards, B)|Cs1], Cs, Gs1, Gs) :-
    delete(Guards, otherwise, Guards2),
    eliminate_guards_otherwise2(NA, Args, Guards2, B, Cs1, Cs, Gs1, Gs).

eliminate_guards_otherwise2(NA, Args, [], B, [], [C], Gs, Gs) :-
    % 'otherwise' with no other guards in last group member - 
    % only force suspension
    C = c(NA, Args, [], ('$suspend', B)).
eliminate_guards_otherwise2(N/A, Args, [], B, Cs1, [C1, C2], Gs, [Cs|Gs]) :-
    % not last, no other guards: force suspension and split group
    build_arglist(0, A, Args2),
    rename_functor(N/A, 1, Nr),
    split_group(Cs1, 1, Cs),
    OCall =.. [Nr|Args2], 
    C1 = c(N/A, Args, [], B),
    C2 = c(N/A, Args2, [], ('$suspend', OCall)).
eliminate_guards_otherwise2(N/A, Args, Guards, B, Cs1, Cs, Gs1, Gs) :-
    % other guards exist, force suspension in wrapper and continue
    build_arglist(0, A, Args2),
    rename_functor(N/A, 1, Nr),
    split_group(Cs1, 1, Cs2),
    FCall =.. [Nr|Args2],
    C1 = c(N/A, Args2, [], ('$suspend', FCall)),
    eliminate_guards2([C1], Cs, [[c(Nr/A, Args, Guards, B)|Cs2]|Gs1], Gs).

eliminate_guards_idle(N/A, Args, Guards, B, Cs1, Cs, Gs1, Gs) :-
    build_arglist(0, A, Args2),
    extract_variables(Args2, Vars),
    available_variable(Vars, IV),
    IVar = '$VAR'(IV),
    append(Args2, [IVar], FParams),
    append(Args2, ['!'(IVar)], FArgs),
    rename_functor(N/A, 1, Nr),
    A2 is A + 1,
    rename_functor(N/A2, 2, i, Nr2),
    split_group(Cs1, 1, Cs2),
    FCall =.. [Nr2|FParams],
    OCall =.. [Nr|Args2], 
    C1 = c(N/A, Args2, [], ('$get_idle'(IVar), FCall)),
    G = [c(Nr2/A2, FArgs, [], OCall)],
    eliminate_guards2([C1], Cs, [G, [c(Nr/A, Args, Guards, B)|Cs2]|Gs1], Gs).

eliminate_guards([Guard|Guards], N/A, Args, Args2, B1, B, Next, Gs1, Gs) :-
    extract_variables(Guard, GuardVars),
    extract_variables((Guard, Args), HeadVars),
    rename_functor(N/A, 2, Nr),
    available_variable(HeadVars, HV),
    ResultVar = '$VAR'(HV),
    translate_guard(Guard, ResultVar, XGuard, [HV|HeadVars], AllVars, Force, Result),
    build_doubler_call(Args, Args2, DoublerParams, CheckArgs, ResultVar, Result, 
        GuardVars, Force, AllVars, DispatchParams, DispatchArgs, MismatchParams,
        MismatchArgs),
    B =.. [Nr|DoublerParams],
    length(CheckArgs, CArity),
    rename_functor(N/A, 3, Nr2),
    DispatchCall =.. [Nr2|DispatchParams],
    Mismatch =.. [Nr2|MismatchParams],
    Gs2 = [[c(Nr/CArity, CheckArgs, [], (XGuard, DispatchCall)),
        c(Nr/CArity, MismatchArgs, [], ('$suspend', Mismatch))], DG|Gs1],
    eliminate_guards2(Next, Guards, Nr2/A, Args, DispatchArgs, Result, B1, DG, 
        Gs2, Gs).

eliminate_guards2([], [], N/A, _, DispatchArgs, _, B, DG, Gs, Gs) :-
    % no other clause or guard: success dispatcher evaluates body
    A2 is A + 1,
    DG = [c(N/A2, DispatchArgs, [], B)].
eliminate_guards2([], Guards, N/A, Args, _, R, B1, DG, Gs1, Gs) :-
    % no other clause, but other guards left: success dispatcher invokes next guard
    A2 is A + 1,
    DG = [c(N/A2, Args3, [], B)],
    eliminate_guards(Guards, N/A, Args, Args2, B1, B, [], Gs1, Gs),
    append(Args2, [R], Args3).
eliminate_guards2(Next, [], N/A, _, DispatchArgs, _, B, DG, Gs, Gs) :-
    % other clauses exist, but no more guards: success evals body, fail calls next
    build_arglist(0, A, Args),
    A2 is A + 1,
    build_arglist(0, A2, Args2),
    NCall =.. [Next|Args],
    DG = [c(N/A, DispatchArgs, [], B), c(N/A, Args2, [], ('$suspend', NCall))].
eliminate_guards2(Next, Guards, N/A, Args, _, R, B1, DG, Gs1, Gs) :-
    % other clauses + guards exist: success calls next guard, fail calls next clause
    A2 is A + 1,
    DG = [c(N/A2, ArgsY, [], B), c(N/A2, ArgsN, [], NCall)],
    eliminate_guards(Guards, N/A, Args, Args2, B1, B, Next, Gs1, Gs),
    append(Args2, [R], ArgsY),
    extract_variables(Args2, Vs),
    available_variable(Vs, V),
    append(Args2, ['$VAR'(V)], ArgsN),
    unforce_arglist(Args2, Params2),
    NCall =.. [Next|Params2].

build_doubler_call([], [], [], [], RV, R, _, _, _, [RV], [R], [NR], []) :-
    opposite(R, NR).
build_doubler_call(['$VAR'(V)|Args], ['!'('$VAR'(V))|DArgs], ['$VAR'(V)|DoP], ['$VAR'(V)|CA], 
        RV, R, GV, yes, HV, ['$VAR'(V)|DiP], ['$VAR'(V)|DA], ['$VAR'(V)|MP], ['$VAR'(V)|MA]) :-
    member(V, GV),
    % forced single variable argument
    build_doubler_call(Args, DArgs, DoP, CA, RV, R, GV, yes, HV, DiP, DA, MP, MA).
build_doubler_call(['$VAR'(V)|Args], ['$VAR'(V)|DArgs], ['$VAR'(V)|DoP], ['$VAR'(V)|CA], RV, 
        R, GV, Force, HV, ['$VAR'(V)|DiP], ['$VAR'(V)|DA], ['$VAR'(V)|MP], ['$VAR'(V)|MA]) :-
    % other single variable argument
    build_doubler_call(Args, DArgs, DoP, CA, RV, R, GV, Force, HV, DiP, DA, MP, MA).
build_doubler_call([X|Args], ['!'('$VAR'(NV))|DArgs], ['$VAR'(NV), '$VAR'(NV)|DoP], 
        [XF, '$VAR'(NV)|CA], RV, R, GV, yes, HV, ['$VAR'(NV)|DiP], [X|DA], 
        ['$VAR'(NV)|MP], ['$VAR'(IV), '$VAR'(NV)|MA]) :-
    refers_to_variables(X, GV),
    % structured argument refering to guard variable, forced
    force_variables(X, GV, XF),
    available_variable(HV, NV),
    available_variable([NV|HV], IV),
    build_doubler_call(Args, DArgs, DoP, CA, RV, R, GV, yes, [IV, NV|HV], DiP, DA, MP, MA).
build_doubler_call([X|Args], ['$VAR'(NV)|DArgs], ['$VAR'(NV), '$VAR'(NV)|DoP], 
        [X, '$VAR'(NV)|CA], RV, R, GV, no, HV, ['$VAR'(NV)|DiP], [X|DA], 
        ['$VAR'(NV)|MP], ['$VAR'(IV), '$VAR'(NV)|MA]) :-
    refers_to_variables(X, GV),
    % structured argument refering to guard variable, non-forced
    available_variable(HV, NV),
    available_variable([NV|HV], IV),
    build_doubler_call(Args, DArgs, DoP, CA, RV, R, GV, no, [IV, NV|HV], DiP, DA, MP, 
        MA).
build_doubler_call([X|Args], ['$VAR'(NV)|DArgs], ['$VAR'(NV)|DoP], ['$VAR'(NV)|CA], RV, 
        R, GV, Force, HV, ['$VAR'(NV)|DiP], [X|DA], ['$VAR'(NV)|MP], ['$VAR'(NV)|MA]) :-
    % structured argument not refering to guard variable
    available_variable(HV, NV),
    build_doubler_call(Args, DArgs, DoP, CA, RV, R, GV, Force, [NV|HV], DiP, DA, MP, MA).

% utilities

split_group([], _, []).
split_group([c(N/A, Args, Gs, B)|G1], P, [c(Nr/A, Args, Gs, B)|G]) :-
    rename_functor(N/A, P, Nr),
    split_group(G1, P, G).

rename_functor(N/A, P, RN) :-
    format_to_atom(RN, '$~d$~a$~d', [P, N, A]).

rename_functor(N/A, P, Grist, RN) :-
    format_to_atom(RN, '$~d$~a$~a$~d', [P, Grist, N, A]).

build_arglist(N, M, L) :- build_arglist(N, M, L, []).

build_arglist(N, N, T, T) :- !.
build_arglist(N, M, ['$VAR'(N)|Args], T) :-
    N < M,
    N2 is N + 1,
    build_arglist(N2, M, Args, T).

build_arglist_excluding(N, N, _, []) :- !.
build_arglist_excluding(N, M, Vs, ['$VAR'(V)|L]) :-
    N < M,
    available_variable(Vs, V),
    N2 is N + 1,
    build_arglist_excluding(N2, M, [V|Vs], L).

extract_variables(X, Vs) :- extract_variables(X, Vs, []).

extract_variables(X, Vs, Vs) :- atomic(X).
extract_variables('$VAR'(V), [V|T], T).
extract_variables([X|Y], Vs, T) :-
    extract_variables(X, Vs, Vs2),
    extract_variables(Y, Vs2, T).
extract_variables(X, Vs, T) :-
    callable(X),
    X =.. [_|Args],
    extract_variables(Args, Vs, T).

refers_to_variables(X, _) :- atomic(X), !, fail.
refers_to_variables('$VAR'(V), Vars) :- member(V, Vars), !.
refers_to_variables('$VAR'(_), _) :- !, fail.
refers_to_variables([X|_], Vars) :- refers_to_variables(X, Vars).
refers_to_variables([_|X], Vars) :- !, refers_to_variables(X, Vars).
refers_to_variables(X, Vars) :- 
    callable(X),
    X =.. [_|Args],
    refers_to_variables(Args, Vars).

force_variables(X, _, X) :- atomic(X).
force_variables('!'(X), _, '!'(X)).
force_variables('$VAR'(V), Vars, '!'('$VAR'(V))) :- member(V, Vars), !.
force_variables('$VAR'(V), _, '$VAR'(V)).
force_variables([X|Y], Vars, [X2|Y2]) :- 
    force_variables(X, Vars, X2),
    force_variables(Y, Vars, Y2).
force_variables(X, Vars, X2) :-
    callable(X),
    X =.. [N|Args],
    force_variables(Args, Vars, Args2),
    X2 =.. [N|Args2].

available_variable(Vars, V) :- available_variable(0, Vars, V).

available_variable(V, Vars, V) :- \+member(V, Vars), !.
available_variable(V1, Vars, V) :-
    V2 is V1 + 1,
    available_variable(V2, Vars, V).

% copy argument list, removing "!"
unforce_arglist([], []).
unforce_arglist(['!'(X)|L], [X|L2]) :- unforce_arglist(L, L2).
unforce_arglist([X|L], [X|L2]) :- unforce_arglist(L, L2).

% Guard translation

translate_guard(Exp, R, (X2, Y2, T), HV1, HV, Force, Result) :-
    callable(Exp),
    functor(Exp, Op, Arity),
    numeric_guard(Op/Arity),
    Exp =.. [_, X, Y],
    translate_expression(X, HV1, X2, XR, HV2),
    translate_expression(Y, HV2, Y2, YR, HV),
    Exp2 =.. [Op, XR, YR],
    translate_guard2(Exp2, R, T, Force, Result).
translate_guard(Exp, R, T, HV, HV, Force, Result) :-
    translate_guard2(Exp, R, T, Force, Result).

translate_guard2((X =:= Y), R, compute(=, X, Y, R), no, true).
translate_guard2('\\=:='(X, Y), R, compute(=, X, Y, R), no, false).
translate_guard2((X == Y), R, '$match'(X, Y, R), yes, true).
translate_guard2((X =\= Y), R, '$match'(X, Y, R), yes, false).
translate_guard2((X > Y), R, call('$rt':greater(X, Y, R)), yes, true).
translate_guard2((X < Y), R, compute(<, X, Y, R), yes, true).
translate_guard2((X >= Y), R, compute(<, X, Y, R), yes, false).
translate_guard2((X =< Y), R, call('$rt':greater(X, Y, R)), yes, false).
translate_guard2((X @< Y), R, compute(@<, X, Y, R), yes, true).
translate_guard2((X @> Y), R, compute(@>, X, Y, R), yes, true).
translate_guard2((X @>= Y), R, compute(@<, X, Y, R), yes, false).
translate_guard2((X @=< Y), R, compute(@>, X, Y, R), yes, false).
translate_guard2(integer(X), R, '$integer'(X, R), yes, true).
translate_guard2(float(X), R, '$float'(X, R), yes, true).
translate_guard2(number(X), R, '$number'(X, R), yes, true).
translate_guard2(string(X), R, '$atom'(X, R), yes, true).
translate_guard2(atomic(X), R, '$atomic'(X, R), yes, true).
translate_guard2(list(X), R, '$list'(X, R), yes, true).
translate_guard2(port(X), R, '$port'(X, R), yes, true).
translate_guard2(tuple(X), R, '$tuple'(X, R), yes, true).
translate_guard2(unknown(X), R, '$var'(X, R), no, true).
translate_guard2(known(X), R, '$var'(X, R), no, false).
translate_guard2(module(X), R, '$module'(X, R), yes, true).
translate_guard2(G, _, _, _, _) :- error('invalid guard: ~q\n', [G]).

opposite(false, true).
opposite(true, false).

translate_expression(rnd, HV, foreign_call(fl_random('$VAR'(V))), '$VAR'(V), [V|HV]) :- 
    !,
    available_variable(HV, V).
translate_expression(Exp, HV, true, Exp, HV) :- atomic(Exp).
translate_expression('$VAR'(V), HV, true, '$VAR'(V), HV).
translate_expression(-X, HV1, X2, XR, HV) :-
    translate_expression(0-X, HV1, X2, XR, HV).
translate_expression(+X, HV1, X2, XR, HV) :-
    translate_expression(X, HV1, X2, XR, HV).
translate_expression(\X, HV1, X2, XR, HV) :-
    translate_expression('><'(-1, X), HV1, X2, XR, HV).
translate_expression(Exp, HV1, (TX, TY, compute(NOp, RX, RY, '$VAR'(V))), '$VAR'(V), [V|HV3]) :-
    Exp =.. [Op, X, Y],
    numeric_operation(Op, NOp),
    translate_expression(X, HV1, TX, RX, HV2),
    translate_expression(Y, HV2, TY, RY, HV3),
    available_variable(HV3, V).
translate_expression(Exp, HV1, (TX, compute(NOp, RX, '$VAR'(V))), '$VAR'(V), [V|HV2]) :-
    Exp =.. [Op, X],
    numeric_operation(Op, NOp),
    translate_expression(X, HV1, TX, RX, HV2),
    available_variable(HV2, V).
translate_expression(Exp, HV, true, Exp, HV).

% expand is/2 expressions, explicit module calls and builtins:

expand_terms([], []).
expand_terms([G1|Gs1], [G|Gs]) :-
    expand_terms2(G1, G, Gs1, Gs2),
    expand_terms(Gs2, Gs).

expand_terms2([], [], Gs, Gs).
expand_terms2([C1|G1], [c(NA, Args, Guards, Body)|G], Gs1, Gs) :-
    C1 = c(NA, Args, Guards, Body1),
    extract_variables((Args, Guards, Body1), Vars),
    expand_body(Body1, Body, NA, Vars, _, Gs1, Gs2),
    expand_terms2(G1, G, Gs2, Gs).

counter(N) :-
    g_read(counter, N1),
    N is N1 + 1, 
    g_assign(counter, N).

expand_body((true, X), X2, NA, Vars1, Vars, Gs1, Gs) :-
    expand_body(X, X2, NA, Vars1, Vars, Gs1, Gs).
expand_body((X, true), X2, NA, Vars1, Vars, Gs1, Gs) :-
    expand_body(X, X2, NA, Vars1, Vars, Gs1, Gs).
expand_body('@'(X, Y), X2, NA, Vars1, Vars, Gs1, Gs) :-
    expand_body(rcall(X, Y), X2, NA, Vars1, Vars, Gs1, Gs).
expand_body(X = Y, X2, NA, Vars1, Vars, Gs1, Gs) :-
    expand_body(foreign_call(fl_unify(X, Y)), X2, NA, Vars1, Vars, Gs1, Gs).
expand_body(X is Exp, T, NA, Vars, [RV|Vars], Gs1, [G|Gs1]) :-
    counter(C),
    rename_functor(NA, C, x, N2),
    extract_variables(Exp, ExpVars),
    sort(ExpVars, SortedVars),
    length(SortedVars, SL),
    available_variable(Vars, RV),
    translate_expression(Exp, [RV|SortedVars], XBody, Result, _),
    build_exp_args(SortedVars, Args, Params, ['$VAR'(RV)]),
    Call =.. [N2|Params],
    T = (Call, (X = '$VAR'(RV))),
    SL2 is SL + 1,
    G = [c(N2/SL2, Args, [], (XBody, '$assign'(Result, '$VAR'(RV))))].
expand_body(when(X, T), T2, NA, Vars, Vars, Gs1, [G|Gs1]) :-
    check_when(X, T, V),
    counter(C),
    rename_functor(NA, C, w, N2),
    extract_variables((X, T), WVars),
    sort(WVars, SortedVars),
    length(SortedVars, SL),
    build_when_args(SortedVars, V, Args, Params),
    T2 =.. [N2|Params],
    G = [c(N2/SL, Args, [], T)].
expand_body(M:T, X, NA, Vars1, Vars, Gs1, Gs) :-
    expand_body(call(M:T), X, NA, Vars1, Vars, Gs1, Gs).
expand_body(apply(T, Args), X, NA, Vars1, Vars, Gs1, Gs) :-
    expand_body(lib:'$apply'(T, Args, '$VAR'('$MODULE')), X, NA, Vars1, Vars, Gs1, Gs).
expand_body((X, Y), (X2, Y2), NA, Vars1, Vars, Gs1, Gs) :-
    expand_body(X, X2, NA, Vars1, Vars2, Gs1, Gs2),
    expand_body(Y, Y2, NA, Vars2, Vars, Gs2, Gs).
expand_body('$VAR'(V), X, NA, Vars1, Vars, Gs1, Gs) :-
    expand_body(call('$VAR'(V)), X, NA, Vars1, Vars, Gs1, Gs).
expand_body(T, T2, NA, Vars, Vars, Gs1, Gs) :-
    callable(T),
    functor(T, NB, A),
    T =.. [_|Args],
    foreign(NB/A, Op, Spec),
    expand_builtin_call(Spec, Args, Op, NA, T2, Gs1, Gs).
expand_body(T, T2, _, Vars, Vars, Gs, Gs) :-
    callable(T),
    functor(T, NB, A),
    \+library_mode,
    T =.. [_|Args],
    library(NB/A, M:Call),
    C =.. [Call|Args],
    T2 = call(M:C).
expand_body(T, T, _, Vs, Vs, Gs, Gs).

check_when('$VAR'(V), _, V).
check_when(X, T, _) :-
    error('invalid argument to `when\': ~q\n', [when(X, T)]).

build_when_args([], _, [], []).
build_when_args([V|Vs], V, ['!'('$VAR'(V))|Args], ['$VAR'(V)|Params]) :-
    build_when_args(Vs, V, Args, Params).
build_when_args([V1|Vs], V, ['$VAR'(V1)|Args], ['$VAR'(V1)|Params]) :-
    build_when_args(Vs, V, Args, Params).

expand_builtin_call(none, Args, Op, _, R, Gs, Gs) :-
    T =.. [Op|Args],
    R = foreign_call(T).
expand_builtin_call(Spec, Args, Op, NA, R, Gs, [G|Gs]) :-
    counter(C),
    rename_functor(NA, C, f, N2),
    extract_variables(Args, Vars),
    sort(Vars, SortedVars),
    length(SortedVars, SL),
    scan_builtin_args(Spec, Args, Forced),
    var_lists(SortedVars, Forced, Args2, Params),
    T =.. [Op|Args],
    R =.. [N2|Params],
    G = [c(N2/SL, Args2, [], foreign_call(T))].

scan_builtin_args([], _, []).
scan_builtin_args([f|Spec], ['$VAR'(V)|Args], [V|Forced]) :-
    scan_builtin_args(Spec, Args, Forced).
scan_builtin_args([_|Spec], [_|Args], Forced) :-
    scan_builtin_args(Spec, Args, Forced).

var_lists([], _, [], []).
var_lists([V|Vs], Forced, ['!'('$VAR'(V))|Args], ['$VAR'(V)|Params]) :-
    member(V, Forced),
    var_lists(Vs, Forced, Args, Params).
var_lists([V|Vs], Forced, ['$VAR'(V)|Args], ['$VAR'(V)|Params]) :-
    var_lists(Vs, Forced, Args, Params).

build_exp_args([], T, T, T).
build_exp_args([V|Vs], ['!'('$VAR'(V))|XVs], ['$VAR'(V)|PVs], T) :-
    build_exp_args(Vs, XVs, PVs, T).

foreign(command_line/1, fl_command_line, [x]).
foreign(string_to_integer/2, fl_string_to_integer, [f, x]).
foreign(string_to_integer/3, fl_string_to_integer, [f, f, x]).
foreign(string_to_float/2, fl_string_to_float, [f, x]).
foreign(string_to_list/3, fl_string_to_list, [f, x, x]).
foreign(string_to_byte_list/3, fl_string_to_byte_list, [f, x, x]).
foreign(number_to_list/3, fl_number_to_list, [f, x, x]).
foreign(number_to_list/4, fl_number_to_list, [f, f, x, x]).
foreign(close_file/2, fl_close_file, [f, x]).
foreign(utf_encode/3, fl_utf_encode, [f, x, x]).
foreign(open_port/2, fl_open_port, [x, x]).
foreign(send/2, fl_send, [f, x]).
foreign(send/3, fl_send, [f, x, x]).
foreign(tuple_to_list/3, fl_tuple_to_list, [f, x, x]).
foreign(put_arg/3, fl_put_arg, [f, f, x]).
foreign(put_arg/4, fl_put_arg, [f, f, x, x]).
foreign(get_arg/3, fl_get_arg, [f, f, x]).
foreign(get_arg/4, fl_get_arg, [f, f, x, x]).
foreign(make_tuple/2, fl_make_tuple, [f, x]).
foreign(halt/1, fl_halt, [f]).
foreign(assign/3, fl_assign, [x, x, x]).
foreign(listen/2, fl_listen, [f, x]).
foreign(threads/1, fl_threads, [x]).
foreign(getpid/1, fl_getpid, [x]).
foreign(self/1, fl_self, [x]).
foreign(current_seconds/1, fl_current_seconds, [x]).
foreign(signal/3, fl_signal, [f, x, x]).
foreign(timer/3, fl_timer, [f, x, x]).
foreign(clock/3, fl_clock, [f, x, x]).
foreign(cancel_timer/2, fl_cancel_timer, [f, x]).
foreign(randomize/2, fl_set_random_seed, [x, x]).
foreign(file_exists/2, fl_file_exists, [f, x]).
foreign(getcwd/1, fl_getcwd, [x]).
foreign(get_global/2, fl_get_global, [f, x]).
foreign(put_global/3, fl_put_global, [f, x, x]).
foreign(compare/3, fl_compare, [f, f, x]).

library(length/2, lib:length).
library(deref/2, lib:deref).
library(error/1, lib:error).
library(writeln/1, lib:writeln).
library(writeln/2, lib:writeln).
library(list_to_string/2, lib:list_to_string).
library(list_to_tuple/2, lib:list_to_tuple).
library(list_to_number/2, lib:list_to_number).
library(list_to_number/3, lib:list_to_number).
library(write_file/3, lib:write_file).
library(write_file/4, lib:write_file).
library(read_file/4, lib:read_file).
library(utf_encode/2, lib:utf_encode).
library(utf_encode/3, lib:utf_encode).
library(utf_decode/2, lib:utf_decode).
library(utf_decode/3, lib:utf_decode).
library(append/3, lib:append).
library(append/2, lib:append).
library(member/3, lib:member).
library(read_byte_stream/2, lib:read_byte_stream).
library(read_char_stream/2, lib:read_char_stream).
library(write_byte_stream/2, lib:write_byte_stream).
library(write_char_stream/2, lib:write_char_stream).
library(log/1, lib:log).
library(get_module/2, lib:get_module).
library(chdir/2, lib:chdir).
library(getenv/2, lib:getenv).
library(file_type/2, lib:file_type).
library(readlink/2, lib:readlink).
library(open_file/3, lib:open_file).
library(delete_file/2, lib:delete_file).
library(rmdir/2, lib:rmdir).
library(chmod/3, lib:chmod).

numeric_guard(('=:=')/2).
numeric_guard(('\\=:=')/2).
numeric_guard(('>')/2).
numeric_guard(('<')/2).
numeric_guard(('>=')/2).
numeric_guard(('=<')/2).

numeric_operation('/\\', and).
numeric_operation('\\/', or).
numeric_operation('><', xor).
numeric_operation('<<', shl).
numeric_operation('>>', shr).
numeric_operation('\\\\', mod).
numeric_operation(Op, Op).

numeric_operation(+).
numeric_operation(-).
numeric_operation(*).
numeric_operation(/).
numeric_operation(sqrt).
numeric_operation(integer).
numeric_operation(float).
numeric_operation(floor).
numeric_operation(ceiling).
numeric_operation(round).
numeric_operation(truncate).
numeric_operation(max).
numeric_operation(min).
